import json
from abc import ABC

import requests

from .unit import UnitTestCase
from ..service.v1.caller import APICaller


class ServiceTestCase(UnitTestCase, ABC, APICaller):
    api_url_base = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.api_url_base = ServiceTestCase.config.get('service', {}).get('port', 80)

    @classmethod
    def request_api(cls, api_url, timeout=5, **kwargs):
        """Class method to test an API call

        :param api_url: do NOT starts with slash (/)
        :param timeout: requests timeout in seconds
        :param kwargs: request data
        :return:
        """
        payload = cls.wrap_request_data(data=kwargs)
        url = 'http://localhost:%s/%s' % (cls.api_url_base, api_url)
        cls.LOG.debug("POST %s  %s" % (url, json.dumps(payload, ensure_ascii=False, sort_keys=True)))
        resp = requests.post(
            url, json=payload, timeout=timeout, headers={'Content-Type': 'application/json'}
        ).json()
        cls.LOG.debug(resp)
        return resp



from secrets import SystemRandom


def random_bool():
    return SystemRandom().choice([True, False])


def random_choice(*args):
    return SystemRandom().choice(*args)


def random_int(a, b):
    return SystemRandom().randint(a, b)


def random_ratio():
    return SystemRandom().random()


def random_uniform(*args):
    return SystemRandom().uniform(*args)


def random_sample(*args):
    return SystemRandom().sample(*args)



import json
import typing
import confluent_kafka as kafka
import confluent_kafka.admin as kafka_admin

from ..logger import LOG

LOG.debug('Version of confluent_kafka client = %s' % kafka.__version__)


class KafkaOperator:
    def __init__(self, kafka_config):
        """
        Parameter reference: https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md

        :param kafka_config:
        host = [
            {host: kafka_server_1, port: 9092}
        ]
        """
        self._config = json.loads(json.dumps(kafka_config, ensure_ascii=False))  # deep copy

        if 'host' in kafka_config:
            self._config = {
                'bootstrap.servers': ','.join(['{host}:{port}'.format(**i) for i in kafka_config.pop('host')]),
            }
        LOG.debug("Kafka connection info: " + str(self._config))

    def admin_client(self, *args, **kwargs):
        config_admin = {**self._config}
        a = kafka_admin.AdminClient(config_admin)
        return a

    def create_topic(self, topic: str, num_partitions=3, replication_factor=1, *args, **kwargs):
        """Note: In a multi-cluster production scenario, it is more typical to use a replication_factor of 3 for durability."""
        a = self.admin_client()
        new_topic = kafka_admin.NewTopic(topic, num_partitions=num_partitions, replication_factor=replication_factor)

        # Call create_topics to asynchronously create topics. A dict of <topic,future> is returned.
        fs = a.create_topics([new_topic])

        # Wait for each operation to finish.
        for topic, f in fs.items():
            try:
                f.result()  # The result itself is None
                LOG.info("Topic {} created".format(topic))
            except Exception as e:
                LOG.error("Failed to create topic {}: {}".format(topic, e))
                return False
            finally:
                a.close()

        return True

    def producer_deliver(self, topic: str, generator: typing.Iterator[str], func_callback: callable = None, *args, **kwargs):
        # func_callback should be a function that takes two arguments: err and msg
        config_producer = {**self._config}
        p = kafka.Producer(config_producer)

        def delivery_report(err, msg):
            """ Called once for each message produced to indicate delivery result. Triggered by poll() or flush(). """
            if err is not None:
                LOG.error('Kafka msg delivery failed: {}'.format(err))
            else:
                LOG.debug('Kafka msg delivered to {} [{}]'.format(msg.topic(), msg.partition()))

        if func_callback is None:
            func_callback = delivery_report

        for data in generator:  # some data from the generator
            # Trigger any available delivery report callbacks from previous produce() calls
            p.poll(0)

            # Asynchronously produce a message, the delivery report callback
            # will be triggered from poll() above, or flush() below, when the message has
            # been successfully delivered or failed permanently.
            p.produce(topic, data.encode('utf-8'), callback=func_callback)

        # Wait for any outstanding messages to be delivered and delivery report callbacks to be triggered.
        p.flush()

    def consumer_generator(self, topics_subscribe: list, group_id: str = None, poll_timeout: float = 1.0, *args, **kwargs) -> typing.Iterator[str]:
        config_consumer = {'auto.offset.reset': 'earliest', **self._config}
        if group_id is not None:
            config_consumer['group.id'] = group_id
        c = kafka.Consumer(config_consumer)

        c.subscribe(topics_subscribe)
        while True:
            msg = c.poll(poll_timeout)

            if msg is None:
                continue
            elif msg.error():
                code = msg.error().code()
                if code == kafka.KafkaError._PARTITION_EOF:
                    pass
                LOG.error("Kafka consumer: {}".format(msg.error()))
                continue

            data = msg.value().decode('utf-8')
            LOG.debug('Received message: {}'.format(data))
            yield data

        c.close()



"""Easily put time restrictions on things
Note: Requires Python 3.x
Usage as a context manager:
```
with TimeOutRestriction(10):
    something_that_should_not_exceed_ten_seconds()
```
Usage as a decorator:
```
@TimeOutRestriction(10)
def something_that_should_not_exceed_ten_seconds():
    do_stuff_with_a_timeout()
```
Handle timeouts:
```
try:
   with TimeOutRestriction(10):
       something_that_should_not_exceed_ten_seconds()
   except TimeoutError:
       log('Got a timeout, couldn't finish')
```
Suppress TimeoutError and just die after expiration:
```
with TimeOutRestriction(10, suppress_timeout_errors=True):
    something_that_should_not_exceed_ten_seconds()
print('Maybe exceeded 10 seconds, but finished either way')
```
"""
import contextlib
import errno
import os
import signal

DEFAULT_TIMEOUT_MESSAGE = os.strerror(errno.ETIME)


class TimeOutRestriction(contextlib.ContextDecorator):
    def __init__(self, milliseconds: float, *, timeout_message: str = DEFAULT_TIMEOUT_MESSAGE, suppress_timeout_errors: bool = False):
        self.millisecond = milliseconds
        self.timeout_message = timeout_message
        self.suppress = bool(suppress_timeout_errors)

    def _timeout_handler(self, signum, frame):
        raise TimeoutError(self.timeout_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._timeout_handler)
        signal.setitimer(signal.ITIMER_REAL, self.millisecond / 1000)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.setitimer(signal.ITIMER_REAL, 0, 0)
        if self.suppress and exc_type is TimeoutError:
            return True



import os

from .config.hocon import load_config_from_hocon
from .config.paths import get_resource_dir, get_config_dir


class Settings:
    def __init__(self):
        self._config = None

    @property
    def resource_dir(self):
        return get_resource_dir()

    @property
    def config_dir(self):
        return get_config_dir()

    @property
    def config(self):
        if self._config is None:
            # load config items from hocon config files
            _file_config_main = os.path.join(self.config_dir, 'main.conf')
            self._config = load_config_from_hocon(_file_config_main)
        return self._config

    def __getitem__(self, item):
        return self.config[item]


SETTINGS = Settings()



import json
from datetime import datetime

from pandas._libs.tslibs.nattype import NaTType
from pandas._libs.tslibs.timestamps import Timestamp


class ObjectWithDateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, NaTType):  # notice: NaTType is a subclass of datetime
            return None
        if isinstance(obj, Timestamp):
            return obj.timestamp()
        if isinstance(obj, datetime):
            return obj.timestamp()

        return json.JSONEncoder.default(self, obj)



import pymysql
from sqlalchemy import create_engine
from sqlalchemy.sql import text

from .base import password_vault
from ..logger import LOG

LOG.debug('Version of pymysql = %s' % pymysql.__version__)


class MySqlOperator:
    def __init__(self, db_config, **kwargs):
        self._config = {
            'user': db_config['user'],
            'password': password_vault.get_password(db_config['password']),
            'host': db_config['host'],
            'port': db_config['port'],
            'dbname': db_config['dbname']
        }

        try:
            self.db = create_engine(
                'mysql+pymysql://{user}:{password}@{host}:{port}/{dbname}'.format(**self._config),
                encoding='utf-8', pool_size=50, pool_recycle=500, pool_pre_ping=True, **kwargs
            )
            LOG.debug("MySQL connected: {host}:{port}/{dbname}".format(**self._config))
        except Exception as e:
            LOG.exception(e)
            raise RuntimeError('Failed to connect to MySQL')

    @property
    def connection(self):
        return self.db

    def execute_query(self, sql, *args, **kwargs):
        with self.db.connect() as conn:
            cur = conn.execute(text(sql), *args, **kwargs)
            return cur



__all__ = ('sign_data', 'sign_check')

import json

from ..encrypt.hash import get_md5_of_str, get_sha256_of_str
from ..settings import SETTINGS

APP_ID_KEYS = SETTINGS.config.get('APP_ID_KEYS', {})

APP_OPTIONS = SETTINGS.config.get('APP_OPTIONS')

FUNC_SIGN_CHECK = {'md5': get_md5_of_str, 'sha256': get_sha256_of_str}

func_sign_check_default = FUNC_SIGN_CHECK.get(APP_OPTIONS.get('sign_method', 'md5'))


def sign_data(salt_uuid: str, app_id: str, app_key: str, data, sign_method: str = None):
    data_str = str(json.dumps(data, ensure_ascii=False, sort_keys=True, separators=(',', ':')))
    public_key = app_id + salt_uuid + data_str + app_key

    func_sign_check = func_sign_check_default if sign_method is None else FUNC_SIGN_CHECK.get(sign_method)
    if func_sign_check is None:
        raise ValueError('Invalid `sign_method`: %s' % sign_method)
    sign = func_sign_check(public_key)
    return sign


def sign_check(salt_uuid: str, app_id: str, sign: str, data, sign_method: str = None, date_time=None):
    """Sign check for request validation
    :param salt_uuid: UUID for sign check and log investigation
    :param app_id: APP ID
    :param sign: Signature = hash( app_id + salt_uuid+ data + app_key )
    :param data: data object, will be serialized to JSON string
    :param date_time: not used for now
    :param sign_method: Sign method, hash function - one of the following: md5, sha256
    :return: if the sign-check has passed.
    """

    func_sign_check = func_sign_check_default if sign_method is None else FUNC_SIGN_CHECK.get(sign_method)
    if func_sign_check is None:
        raise ValueError('Invalid `sign_method`: %s' % sign_method)

    app_key = APP_ID_KEYS.get(app_id)
    if app_key is None:  # APP_ID not in the dict, unknown APP_ID
        return False

    data_str = str(json.dumps(data, ensure_ascii=False, sort_keys=True, separators=(',', ':')))

    public_key = app_id + salt_uuid + data_str + app_key
    right_sign = func_sign_check(public_key)

    return right_sign == sign



import re

from lxml import etree


def extract_img_url(string):
    try:
        if string is None:
            return None
        html = etree.HTML(string)
        for ii in html:
            images = ii.xpath('p/img/@src')
            return images[0]
    except Exception as e:
        print(e, string)


def extract_text(raw_data):
    if raw_data is not None:
        raw_data = re.sub(r"<script>[\s\S]*</script>", "", raw_data)
        html = etree.HTML(raw_data)

        content = []
        if html is not None:
            html_data = html.xpath('/html/body/*//text()')
            for data in html_data:
                tmp = data.strip(' \n\r').replace('\n', '').replace('\t', '').replace(u'\u3000', u'') \
                    .replace(u'\xa0', u'').replace('\r', '').replace(u'\u2028', u'').replace(u'\u2029', u'')
                if tmp:
                    content.append(tmp)

        item_article = ''.join(content)
        return item_article
    else:
        return None



__all__ = ('parser', 'subparsers')

import argparse

_sub_commands = {
    'info': {
        'version': (('-v', '--version'), dict(action='count', default=0, help='display package version information'))
    },
    'compile': {
        'keep': (('-k', '--keep'), dict(action='append', default=[], help='source files keep as is and not converting to dynamic library')),
        'exclude': (('-x', '--exclude'), dict(action='append', default=[], help='list of files/folders to exclude')),
        'dist': (('-d', '--dist'), dict(default='build', help='target directory for build files')),
        'base': (('-b', '--base'), dict(default=None, help='base dir which will be traversed, use current directory by default'))
    }
}

parser = argparse.ArgumentParser()

_subparsers = parser.add_subparsers(dest='cmd', help="aloha sub-command")
subparsers = {}

for sub_command, args in _sub_commands.items():
    s = _subparsers.add_parser(sub_command)
    for a, (flags, params) in args.items():
        s.add_argument(*flags, **params)
    subparsers[sub_command] = s



import ipaddress
import json

import pymongo

from .base import password_vault
from ..logger import LOG


def _is_ip_addr(s):
    try:
        ipaddress.ip_address(s)
        return True
    except ValueError:
        return False


_conn = {}


def MongoOperator(config):
    db_name = config.get('db_name')
    collection_name = config.get('collection_name')

    _config = {k: v for k, v in config.items() if v is not None}
    key = '%s:%s:%s' % (
        json.dumps(_config, sort_keys=True, ensure_ascii=False),
        db_name or '',
        collection_name or ''
    )

    if key not in _conn:
        try:
            _conn[key] = _MongoDBOperation(_config, db_name=db_name, collection_name=collection_name)
        except Exception as e:
            LOG.exception(e)
            return
    return _conn[key]


class _MongoDBOperation:
    def __init__(self, config, db_name=None, collection_name=None):
        self.db_name, self.collection_name = db_name, collection_name

        host = config['host']
        LOG.debug("Mongo connection info: %s", 'host')

        if config.get('port') is None and isinstance(host, list):
            hosts = ['{host}:{port}'.format(**h) for h in host]
        else:
            hosts = ['{host}:{port}'.format(host=host, port=config['port'])]

        replicaSet = config.get('replicaSet')
        if replicaSet is None and not _is_ip_addr(hosts[0].split(':')[0]):
            # if `replicaSet` not defined, and host in config is domain name (not IP)
            replicaSet = hosts[0].split('.')[0]  # use the first segment of domain name as replicaSet

        _config = {
            'host': 'mongodb://%s' % ','.join(hosts),
            'port': config.get('port'),
            'replicaSet': replicaSet,
            'username': config['username'],
            'password': password_vault.get_password(config.get('password')),
            'maxPoolSize': config.get('maxPoolSize'),
            'authSource': config.get('authSource', db_name)
        }

        try:
            self.conn = pymongo.MongoClient(**_config)

            self.db = self.conn[db_name]
            if self.collection_name is not None:
                self.collection = self.db[self.collection_name]
        except Exception as e:
            LOG.exception(e)

    def set_collection(self, collection_name):
        if collection_name not in self.db.list_collection_names():
            raise Exception("Collection[%s] does not exist in [%s]" % (self.collection_name, self.db_name))
        self.collection_name = collection_name
        self.collection = self.db[self.collection_name]
        return True

    def check_and_get_collection(self, collection_name=None, raise_if_not_exists=True):
        """
        检查collection是否存在，如果存在则返回collection对象，否则抛出异常
        Args:
            @param collection_name: str或unicode    collection名称
        Returns:
            @return: collection对象
        Raises:
            @raise Exception: 如果collection在对应db中不存在，则
        """
        self.db = self.conn[self.db_name]

        if self.collection_name is not None:
            self.collection = self.db[self.collection_name]

        if collection_name is not None and collection_name != self.collection_name:
            if self.collection_name not in self.db.list_collection_names():
                if raise_if_not_exists:
                    raise Exception("Collection [%s] does not exist in [%s]" % (self.collection_name, self.db_name))
                else:
                    pass

            self.collection_name = collection_name
            self.collection = self.db[self.collection_name]

        return self.collection

    def insert(self, doc_or_docs, check_keys=False, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            return collection.insert(doc_or_docs, check_keys=check_keys)
        except Exception as e:
            LOG.exception(e)

    def insert_many(self, docs, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            return collection.insert_many(docs)
        except Exception as e:
            LOG.exception(e)

    def insert_one(self, doc, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            return collection.insert_one(doc)
        except Exception as e:
            LOG.exception(e)

    def delete_many(self, field_filter, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            return collection.delete_many(filter=field_filter)
        except Exception as e:
            LOG.exception(e)

    def delete_one(self, field_filter, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            return collection.delete_one(filter=field_filter)
        except Exception as e:
            LOG.exception(e)

    def update_one(self, field_filter, update, upsert=False, bypass_document_validation=False,
                   collation=None, array_filters=None, session=None, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            collection.update_one(filter=field_filter, update=update, upsert=upsert, bypass_document_validation=bypass_document_validation,
                                  collation=collation, array_filters=array_filters, session=session)
            return True
        except Exception as e:
            LOG.exception(e)
            return False

    def update_many(self, field_filter, update, upsert=False, bypass_document_validation=False,
                    collation=None, array_filters=None, session=None, collection_name=None):
        try:
            collection = self.check_and_get_collection(collection_name)
            return collection.update_many(filter=field_filter, update=update, upsert=upsert, bypass_document_validation=bypass_document_validation,
                                          collation=collation, array_filters=array_filters, session=session)
        except Exception as e:
            LOG.exception(e)

    def query(self, field_filter=None, sort=None, limit=40, skip=0, collection_name=None):
        """
        从mongo查询数据
        Args:
            @param field_filter: dict     根据mongo查询语法构造查询条件
            @param sort: array      排序条件，例如：[("company_name",pymongo.ASCENDING), ("_id",pymongo.ASCENDING)]
            @param limit: int       查询条数
            @param skip: int        controls the starting point of the results set
            @param collection_name: str    collection名称，在建立完连接后可动态更换要查询的collection
        Returns:
            @return: 返回查询结果的游标对象
        """
        try:
            collection = self.check_and_get_collection(collection_name)
            if sort:
                result = collection.find(field_filter or {}).sort(sort).limit(limit)
            else:
                result = collection.find(field_filter or {}).skip(skip).limit(limit)
            return result
        except Exception as e:
            LOG.exception(e)

    def find_many(self, field_filter=None, projection=None, collection_name=None, *args, **kwargs):
        """
        从mongo查询数据，返回所有返回数据的游标
        Args:
            @param field_filter: dict     根据mongo查询语法构造查询条件
            @param projection: dict     限制返回的字段 exmaple: { name: 1, contribs: 1, _id: 0 }
            @param collection_name: str    collection名称，在建立完连接后可动态更换要查询的collection
        """
        try:
            collection = self.check_and_get_collection(collection_name)
            result = collection.find(field_filter or {}, projection, *args, **kwargs)
            return result
        except Exception as e:
            LOG.exception(e)

    def find_one(self, field_filter=None, projection=None, collection_name=None, *args, **kwargs):
        """
        从mongo查询数据,只返回单个结果
        Args:
            @param field_filter: dict     根据mongo查询语法构造查询条件
            @param projection: dict     限制返回的字段 exmaple: { name: 1, contribs: 1, _id: 0 }
            @param collection_name: str    collection名称，在建立完连接后可动态更换要查询的collection
        """
        try:
            collection = self.check_and_get_collection(collection_name)
            result = collection.find_one(field_filter or {}, projection, *args, **kwargs)
            return result
        except Exception as e:
            LOG.exception(e)

    def count(self, field_filter=None, collection_name=None):
        """
        从mongo查询数据条数
        Args:
            @param field_filter: dict    根据mongo查询语法构造查询条件
            @param collection_name: str    collection名称，在建立完连接后可动态更换要查询的collection
        """
        try:
            collection = self.check_and_get_collection(collection_name)
            result = collection.count_documents(field_filter or {})
            return result
        except Exception as e:
            LOG.exception(e)

    def check_connected(self):
        if not self.conn.connected:
            raise NameError('stat:connected Error')



import json
from datetime import datetime, timedelta
from typing import Optional

import requests

from ...logger import LOG

try:
    from simplejson.errors import JSONDecodeError
except ImportError:
    from json import JSONDecodeError


class EsgClient:
    def __init__(self, url_oauth_get_token: str, client_id: str, client_secret: str, grant_type: str = 'client_credentials'):
        self.url_oauth_get_token = url_oauth_get_token
        self.client_id = client_id
        self.client_secret = client_secret
        self.grant_type = grant_type

        self.expires_at = None
        self.access_token = None

    def get_access_token(self) -> str:
        now = datetime.now()

        if self.expires_at is None or self.expires_at > now:
            try:
                # refresh access_token
                resp = requests.post(self.url_oauth_get_token, timeout=5, json={
                    'client_id': self.client_id,
                    'client_secret': self.client_secret,
                    'grant_type': self.grant_type
                })

                data = resp.json()['data']
                if data is None or 'access_token' not in data:
                    raise RuntimeError('Fail to fetch OpenAPI token with result: %s' % resp.text)

                self.access_token = data['access_token']

                expires_in = int(data['expires_in'])
                self.expires_at = datetime.now() + timedelta(minutes=expires_in - 1)
            except Exception as e:
                LOG.error('Exception acquiring ESG access token from [%s]: %s' % (self.url_oauth_get_token, str(e)))

        return self.access_token

    def _get_request_url(self, url: str):
        request_url = '{url}?access_token={access_token}&request_id={request_id}'.format(
            url=url, access_token=self.get_access_token(), request_id=datetime.now().strftime('%Y%m%d-%H%M%S-%f')
        )
        return request_url

    @staticmethod
    def _get_data_from_esg_response(resp) -> Optional[dict]:
        try:
            return resp.json()
        except (json.JSONDecodeError, JSONDecodeError):  # requests may use `simplejson`
            try:
                # when data is wrapped by ESG
                content = resp.text.replace('"data":"', '"data":').replace('}"}', '}}')
                data = json.loads(content)
                return data.get('data', {})
            except json.JSONDecodeError:
                msg = 'Cannot parse ESG response: %s' % resp.text
                raise ValueError(msg)

    def post(self, url_api: str, body: dict, headers: dict = None, timeout: int = 5):
        url = self._get_request_url(url_api)
        LOG.debug('Calling ESG POST: %s' % url)
        try:
            resp = requests.post(url=url, headers=headers, json=body, timeout=timeout)
            return self._get_data_from_esg_response(resp)
        except Exception as e:
            LOG.error('Error calling ESG API [%s]: %s' % (url, str(e)))

    def get(self, url_api: str, body: dict, headers: dict = None, timeout: int = 5):
        url = self._get_request_url(url_api)
        LOG.debug('Calling ESG GET: %s' % url)
        try:
            resp = requests.get(url=url, headers=headers, json=body, timeout=timeout)
            return self._get_data_from_esg_response(resp)
        except Exception as e:
            LOG.error('Error calling ESG API [%s]: %s' % (url, str(e)))



import psycopg2
from sqlalchemy import create_engine
from sqlalchemy.sql import text

from .base import password_vault
from ..logger import LOG

LOG.debug('postgres: psycopg2 version = %s' % psycopg2.__version__)


class PostgresOperator:
    def __init__(self, db_config, **kwargs):
        self._config = {
            'user': db_config['user'],
            'password': password_vault.get_password(db_config.get('password')),
            'host': db_config['host'],
            'port': db_config['port'],
            'dbname': db_config['dbname']
        }
        connect_args = {}
        if 'schema' in db_config:
            connect_args['options'] = '-csearch_path={}'.format(db_config['schema'])

        try:
            self.engine = create_engine(
                'postgresql+psycopg2://{user}:{password}@{host}:{port}/{dbname}'.format(**self._config),
                connect_args=connect_args, client_encoding='utf8', encoding='utf-8',
                pool_size=20, max_overflow=10, pool_pre_ping=True, **kwargs
            )
            LOG.debug("PostgresSQL connected: {host}:{port}/{dbname}".format(**self._config))
        except Exception as e:
            LOG.error(e)
            raise RuntimeError('Failed to connect to PostgresSQL')

    @property
    def connection(self):
        return self.engine

    def execute_query(self, sql, *args, **kwargs):
        with self.engine.connect() as conn:
            cur = conn.execute(text(sql), *args, **kwargs)
            return cur



from abc import ABC

from .unit import UnitTestCase
from ..service.v2.caller import APICaller


class ServiceTestCase(UnitTestCase, ABC, APICaller):
    api_url_base = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.api_url_port = ServiceTestCase.config.get('service', {}).get('port', 80)

    @classmethod
    def request_api(cls, api_url, timeout=5, **kwargs):
        """Class method to test an API call

        :param api_url: do NOT start with slash (/)
        :param timeout: requests timeout in seconds
        :param kwargs: request data
        :return:
        """
        url = 'http://localhost:%s/%s' % (cls.api_url_port, api_url)
        # cls.LOG.debug("POST %s  %s" % (url, json.dumps(kwargs, ensure_ascii=False, sort_keys=True)))
        # resp = requests.post(
        #     url, json=kwargs, timeout=timeout, headers={'Content-Type': 'application/json'}
        # ).json()
        # cls.LOG.debug(resp)
        return cls.call(url, timeout=timeout, **kwargs)



from abc import ABC, abstractmethod
from urllib.parse import quote_plus as urlquote


class BaseVault(ABC):
    @abstractmethod
    def decrypt_password(self, *args, **kwargs):
        return kwargs.get('password')

    def get_password(self, *args, **kwargs):
        url_quote = kwargs.pop('url_quote', True)

        pwd = self.decrypt_password(*args, **kwargs)
        if pwd is None:
            return None

        if url_quote:
            return urlquote(pwd)
        else:
            return pwd


class DummyVault(BaseVault):
    def decrypt_password(self, password):
        return password


def main():
    vault = DummyVault()
    ret = vault.get_password(None, url_quote=True)
    # print(ret)
    return ret



from ..logger import LOG
from ..settings import SETTINGS


def _get_vault():
    encryption_method = SETTINGS.config.get('PASSWORD_ENCRYPTION')
    LOG.info('Using password vault: %s', encryption_method)

    if encryption_method == 'plain' or encryption_method is True:
        from ..encrypt.vault.plain import PlainVault
        return PlainVault()
    elif encryption_method == 'cyberark':
        config_cyberark = SETTINGS.config.get('CYBERARK_CONFIG')
        if config_cyberark is None:
            LOG.error('Missing [CYBERARK_CONFIG] in config!')
            exit(-1)

        from ..encrypt.vault.cyberark import CyberArkVault
        return CyberArkVault(**config_cyberark)

    LOG.info('Using plain password vault as unknown value of PASSWORD_ENCRYPTION=%s in config.', encryption_method)

    from ..encrypt.vault.base import DummyVault
    return DummyVault()


password_vault = _get_vault()



import json

from elasticsearch import helpers
from elasticsearch import Elasticsearch

from .base import password_vault
from ..logger import LOG


class ElasticSearchOperator:
    def __init__(self, config, index_config=None):
        self.es_config = config

        username = config.get('username')
        password = password_vault.get_password(config.get('password'))
        #password = config.get('password')
        self._config = {
            'http_auth': (username, password) if username is not None and password is not None else None,
            'hosts': config.get('host', 'localhost'),

            'timeout': config.get("timeout", 0.1),
            'max_retries': config.get("max_retries", 3),
            'retry_on_timeout': config.get("retry_on_timeout", True)
        }
        LOG.debug("ElasticSearch connection info: " + str(self._config['hosts']))

        self.index_config = index_config
        self.index_name = self.es_config.get('index_name')
        self.index_type = self.es_config.get('index_type')

        self.es = Elasticsearch(**self._config)

        if index_config is not None:
            self.index_config = self._load_config(index_config)

    @staticmethod
    def _load_config(config):
        if type(config) == dict:
            return config

        elif type(config) == str and ".json" in config:
            with open(config, "r", encoding="utf-8") as f:
                config = json.load(f)
            return config

        else:
            raise Exception("Invalid ES config data type")

    def put_mapping(self, index_name=None, index_type=None, index_config=None):
        res = self.es.indices.put_mapping(
            index=index_name or self.index_name,
            doc_type=index_type or self.index_type,
            body=index_config["mappings"][index_type or self.index_type]
        )
        return res

    def build_index(self, index_name=None, index_config=None, raise_if_exist=False):
        if self.es.indices.exists(index=index_name or self.index_name) is not True:
            res = self.es.indices.create(
                index=index_name or self.index_name,
                body=index_config or self.index_config
            )
            return res
        else:
            msg = "Index [%s] already exits" % self.index_name
            if raise_if_exist:
                raise RuntimeError(msg)
            else:
                LOG.info(msg)
                return False

    def delete_index(self, index_name=None, **kwargs):
        res = self.es.indices.delete(index=index_name or self.index_name, ignore=[400, 404])
        return res

    def search(self, query, index_name=None, index_type=None):
        try:
            results = self.es.search(index=index_name or self.index_name, doc_type=index_type or self.index_type, body=query)
            return results
        except Exception as e:
            LOG.exception(e)

    def insert(self, doc, index_name=None, index_type=None, id=None):
        try:
            return self.es.index(index=index_name or self.index_name, doc_type=index_type or self.index_type, id=id, body=doc)
        except Exception as e:
            LOG.exception(e)

    def insert_doc_type(self, **kwargs):
        """
        insert typed json document into index
        :param kwargs:
        :return:
        """
        docs = [[{"create": {"_index": kwargs['index'], "_type": kwargs['doc_type'], "_id": ix}}, doc]
                for ix, doc in enumerate(kwargs['docs'])]
        docs_ = []  # lines:{metadata},content
        [docs_.extend(doc) for doc in docs]
        for i in range(len(docs_))[::kwargs['chunck']]:
            res = self.es.bulk(docs_[i:i + kwargs['chunck']])
            if res['errors']:
                print(res['errors'])
                LOG.exception(res['errors'])

    def insert_update_bulk(self, data):
        helpers.bulk(self.es, data, request_timeout=60)

    def delete_bulk(self, **kwargs):
        """
        delete typed json document into index
        :param kwargs:
        :return:
        """
        docs = [[{"delete": {"_index": kwargs['index'], "_type": kwargs['doc_type'], "_id": _id}}]
                for _id in kwargs['id']]
        docs_ = []  # lines:{metadata},content
        [docs_.extend(doc) for doc in docs]
        for i in range(len(docs_))[::kwargs['chunck']]:
            res = self.es.bulk(docs_[i:i + kwargs['chunck']])
            if res['errors']:
                print(res['errors'])
                LOG.exception(res['errors'])

    def delete_docs(self, **kwargs):
        try:
            return self.es.delete(kwargs['index'], kwargs['doc_type'], kwargs['id'])
        except Exception as e:
            LOG.exception(e)



import hashlib
import requests
from Crypto.Cipher import AES
from binascii import a2b_hex
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from .base import BaseVault
from ...encrypt.aes import AesEncryptor
from ...logger import LOG

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGHT:!DH:!aNULL'


class CyberArkVault(BaseVault, AesEncryptor):
    _cached: dict = {}

    def __init__(self, url: str, app_id: str, key: str = None, safe: str = 'AIM_ELIS_LAS', folder: str = 'root'):
        super().__init__(key)
        self.url = url
        self.app_id = app_id
        self.safe = safe
        self.folder = folder

    @staticmethod
    def get_sign(appid, keyvalue):
        hash_string = appid + "&" + keyvalue
        sha256 = hashlib.sha256()
        sha256.update(hash_string.encode("utf8"))
        return sha256.hexdigest()

    def decrypt_password(self, text):
        if text is None:
            return None

        cryptor = AES.new(self.auto_fill_key(), AES.MODE_ECB)
        s = cryptor.decrypt(a2b_hex(text.encode()))
        s = s[0: -s[-1]]
        return s.decode()

    def get_cyberark_password(self, object=None, **kwargs):
        if isinstance(object, dict):
            kwargs.update(object)
        elif isinstance(object, str):
            kwargs.update({'object': object})

        app_id = kwargs.get('app_id', self.app_id)
        data = {
            "appId": app_id,
            "safe": kwargs.get('safe', self.safe),
            "folder": kwargs.get('folder', self.folder),
            "object": kwargs.get('object', object),
            "sign": self.get_sign(app_id, self.key)
        }

        retry = 5
        while retry:
            try:
                LOG.debug('POST CyberArk: %s with data: %s', self.url, data)
                resp = requests.post(self.url, json=data, verify=False, headers={'Content-Type': 'application/json'})
                tmp = resp.json()
                if resp.status_code == 200 and int(tmp['code']) == 200:
                    LOG.debug('Got data from CyberArk: %s', tmp)
                    return self.decrypt_password(tmp['password'])
                else:
                    raise RuntimeError(resp.text)
            except Exception as e:
                retry -= 1
                if retry == 0:
                    raise e
                else:
                    LOG.error('CyberArk request error: {}'.format(e))
        return None

    def get_password(self, object=None, **kwargs):
        key_for_cache = '{app_id};{safe};{folder};{key};{object}'.format(
            app_id=self.app_id, safe=self.safe, folder=self.folder, key=self.key, object=object
        )
        if key_for_cache not in self._cached:
            pwd = self.get_cyberark_password(object=object, **kwargs)
            self._cached[key_for_cache] = pwd
        else:
            LOG.debug('Using cached CyberArk key: %s' % key_for_cache)

        return self._cached[key_for_cache]


def main():
    vault = CyberArkVault(
        url='https://localhost/pidms/rest/pwd/getPassword',
        app_id='',
        safe='',
        folder='root',
        key='',
    )

    for _ in range(5):
        pwd = vault.get_password({'object': 'DB-*'})
        # print(pwd)



import abc
import json
import logging
from datetime import datetime
from tornado import web
from typing import Optional, Awaitable

from ...encrypt import jwt
from ...logger import LOG
from ...settings import SETTINGS

_RESP_BAD_REQUEST = {'code': '5101', 'message': ['Bad request: fail to parse body as JSON object!']}


class APIHandler(web.RequestHandler):
    LOG = LOG

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def response(self, *args, **kwargs) -> dict:
        raise NotImplementedError()

    def set_default_headers(self) -> None:
        self.set_header('Content-Type', 'application/json; charset=utf-8')

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def on_finish(self) -> None:
        func_callback = getattr(self, 'callback', None)
        if not callable(func_callback):
            return

        body = json.loads(self.request.body.decode('utf-8'))
        return func_callback(**body)

    async def prepare(self, ) -> Optional[Awaitable[None]]:
        access_token = self.request.headers.get('Access-Token')
        if access_token is None:
            return self.finish({
                'msg': 'Invalid Access-Token in request header!'
            })
        else:
            secret_key = SETTINGS.config['APP_SECRET_KEY']  # 'HCTECH-ASKBOB-REC:10062462'
            access_token = jwt.decode(secret_key, access_token)
            if not isinstance(access_token, dict):
                return self.finish({
                    'msg': access_token
                })

    async def get(self, *args, **kwargs):
        query = {k: v[0].decode('utf-8') for k, v in self.request.arguments.items()}
        request_id = self.request.headers.get('Request-ID')
        if request_id is None:
            request_id = datetime.now().strftime('%Y%m%d-%H%M%S-%f')

        try:
            self.LOG.debug('GET Request [%s]: %s' % (request_id, query))
            resp = self.response(**query)  # this call may throw TypeError when argument missing
        except Exception as e:
            self.LOG.error(e, exc_info=True)
            self.LOG.info('POST Request [%s]: %s' % (request_id, query))
            return self.finish({'status': 'error', 'message': [str(e)]})

        resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)

    async def post(self, *args, **kwargs):
        try:
            body = self.request.body.decode('utf-8')
            data = json.loads(body)
        except json.decoder.JSONDecodeError:  # invalid request body, cannot be parsed as JSON
            return self.finish(_RESP_BAD_REQUEST)

        request_id = self.request.headers.get('Request-ID')
        if request_id is None:
            request_id = datetime.now().strftime('%Y%m%d-%H%M%S-%f')

        try:
            if self.LOG.level == logging.DEBUG:
                if len(body) < 1000:
                    self.LOG.debug('POST Request [%s]: %s' % (request_id, data))
                else:  # when body is too long, just print first 1000 char
                    self.LOG.debug('POST Request [%s]: %s' % (request_id, body[:1000]))
            resp = self.response(**data)  # this call may throw TypeError when argument missing
        except Exception as e:
            self.LOG.error(e, exc_info=True)
            self.LOG.info('POST Request [%s]: %s' % (request_id, data))
            return self.finish({'status': 'error', 'message': [str(e)]})

        resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)



import time
from logging import StreamHandler
from logging.handlers import BaseRotatingHandler


class MultiProcessSafeDailyRotatingFileHandler(BaseRotatingHandler):
    """Similar with `logging.TimedRotatingFileHandler`, while this one is
    - Multi process safe
    - Rotate at midnight only
    - Utc not supported
    """

    def __init__(self, filename: str, encoding='utf8', delay=False, utc=False, **kwargs):
        self.utc = utc
        self.suffix = "%Y-%m%d"
        self.baseFilename = filename
        self.currentFileName = self._compute_fn()
        BaseRotatingHandler.__init__(self, filename, 'a', encoding, delay)

    def shouldRollover(self, record):
        if self.currentFileName != self._compute_fn():
            return True
        return False

    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        self.currentFileName = self._compute_fn()
        if not self.delay:
            self.stream = self._open()

    def _compute_fn(self):
        return self.baseFilename.replace(".log", "") + "_" \
               + time.strftime(self.suffix, time.localtime()) \
               + '.log'

    def _open(self):
        return open(self.currentFileName, mode=self.mode, encoding=self.encoding)

    def close(self):
        """Closes the stream."""
        self.acquire()
        try:
            try:
                if self.stream:
                    try:
                        self.flush()
                    finally:
                        stream = self.stream
                        self.stream = None
                        if hasattr(stream, "close"):
                            # print('ttt')
                            stream.close()
            finally:
                # Issue #19523: call unconditionally to prevent a handler leak when delay is set
                StreamHandler.close(self)
        finally:
            self.release()



import uuid

import requests

from ..service.param_check import sign_data
from ..settings import SETTINGS

APP_ID_KEYS = SETTINGS.config.get('APP_ID_KEYS', {})


class APICaller:
    @staticmethod
    def wrap_request_data(
            data, app_id: str = None, app_key: str = None, salt_uuid: str = None, sign: str = None, sign_method: str = None
    ):
        if app_id is None:
            if len(APP_ID_KEYS) != 1:
                raise RuntimeError('Please specify 1 and only 1 in APP_ID_KEYS in configurations!')
            app_id = list(APP_ID_KEYS.keys())[0]
        salt_uuid = salt_uuid or str(uuid.uuid1())
        sign = sign or sign_data(
            salt_uuid=salt_uuid,
            app_id=app_id,
            app_key=app_key or APP_ID_KEYS.get(app_id),
            data=data,
            sign_method=sign_method
        )
        return {
            'salt_uuid': salt_uuid,
            'app_id': app_id,
            'sign': sign,
            'data': data
        }

    @staticmethod
    def call(api_url, timeout=5, **kwargs):
        """

        :param api_url: do NOT starts with slash (/)
        :param timeout: requests timeout in seconds
        :param kwargs: request data
        :return:
        """
        payload = APICaller.wrap_request_data(data=kwargs)
        resp = requests.post(
            api_url, json=payload, timeout=timeout, headers={'Content-Type': 'application/json'}
        ).json()
        ret = resp['data']
        return ret



import importlib

from .base import parser, subparsers


def main():
    args, _ = parser.parse_known_args()

    cmd = args.cmd
    module = '%s.%s' % (__package__, cmd)
    try:
        module = importlib.import_module(module)
    except ImportError:
        print('Invalid sub-command: %s\n\tFailed to import: %s' % (cmd, module))
        exit(-1)

    func_main = getattr(module, 'main')

    sub_parser = subparsers[cmd]
    kwargs, _ = sub_parser.parse_known_args()
    kwargs = vars(kwargs)

    exit(func_main(**kwargs))


if __name__ == '__main__':
    main()



import asyncio
from functools import wraps, partial

__all__ = ('timeout',)


def aioify(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        p_func = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, p_func)

    return run


def timeout(func, timeout=1):
    async def f():
        try:
            result = await asyncio.wait_for(aioify(func), timeout=timeout)
            return result
        except asyncio.TimeoutError:
            raise TimeoutError()
    return f



from datetime import datetime

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

_now = datetime.now()
_version = '%s.%02d%02d.%02d%02d' % (_now.year, _now.month, _now.day, _now.hour, _now.minute)

with open('./aloha/_version.py', 'wt') as fp:
    fp.write('__version__ = "%s"\n' % _version)

setup(
    name='aloha',
    version=_version,
    author='QPod',
    author_email='45032326+QPod0@users.noreply.github.com',
    license='Apache Software License',
    url='https://github.com/QPod/aloha',
    project_urls={
        'Source': 'https://github.com/QPod/aloha',
        'CI Pipeline': 'https://github.com/QPod/aloha/actions',
        'Documentation': 'https://github.com/QPod/aloha/wiki',
    },

    packages=find_packages(where="."),
    include_package_data=True,
    package_data={},
    platforms='Linux, Mac OS X, Windows',
    zip_safe=False,
    install_requires=[
        'pyhocon', 'pycryptodome'
    ],
    extras_require={
        'build': ['Cython'],
        'service': ['tornado'],
        'db': ['sqlalchemy', 'psycopg2-binary', 'pymysql', 'elasticsearch', 'pymongo', 'redis>4.1.0'],
        'stream': ['confluent_kafka'],
        'data': ['pandas'],
        'report': ['openpyxl>=3']
    },
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'aloha = aloha.script.entry_points:main'
        ]
    },

    data_files=[],
    description='Aloha.',
    long_description=long_description,
    long_description_content_type="text/markdown",

    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ]
)



import requests
import uuid
from datetime import datetime, timedelta

from ...encrypt.jwt import encode
from ...logger import LOG
from ...settings import SETTINGS

APP_ID_KEYS = SETTINGS.config.get('APP_ID_KEYS', {})
APP_SECRET_KEY = SETTINGS.config['APP_SECRET_KEY']


class APICaller:
    @staticmethod
    def get_headers(app_id: str = None, app_key: str = None) -> dict:
        if app_id is None:
            # if len(APP_ID_KEYS) != 1:
            #     raise RuntimeError('Please specify 1 and only 1 in APP_ID_KEYS in configurations!')
            app_id = list(APP_ID_KEYS.keys())[0]

        expire_time = datetime.now() + timedelta(days=1)

        access_token = encode(
            secret_key=APP_SECRET_KEY,
            payload={
                'exp': int(expire_time.timestamp()),
                'aid': app_id
            }
        )

        headers = {
            'Content-Type': 'application/json',
            'Access-Token': access_token,
            'Request-ID': str(uuid.uuid1()),
        }
        return headers

    @classmethod
    def call(cls, api_url, timeout=5, **kwargs) -> str:
        """

        :param api_url: do NOT start with slash (/)
        :param timeout: requests timeout in seconds
        :param kwargs: request data
        :return:
        """
        LOG.debug('Calling API: %s' % api_url)
        # LOG.debug('Param: %s' % json.dumps(kwargs, ensure_ascii=False))
        resp = requests.post(
            api_url,
            json=kwargs,
            timeout=timeout,
            headers=cls.get_headers()
        ).json()
        return resp



#!/usr/bin/env python3
# This script build a given python package into a package of dynamic library (.so) files.

__all__ = ('build', 'main')

import glob
import os
import shutil
import time
from collections import defaultdict
from distutils.core import setup

from Cython.Build import cythonize


def _expand(patterns: list = None):
    files = []
    for pattern in patterns:
        for file in glob.glob(pattern):
            files.append(os.path.abspath(file))
    files = sorted(set(files))
    return files


def build(base: str = None, dist: str = 'build', exclude: list = None, keep: list = None, copy_others=True):
    path_base = base or os.path.abspath('.')
    path_build = dist or os.path.abspath(dist)
    files_exclude = _expand(exclude or [])  # sorted(set(os.path.abspath(i) for i in files_exclude))
    files_keep = _expand(keep or [])  # sorted(set((os.path.abspath(i) for i in files_keep)))

    # decide which files to copy and which files to cythonize
    act_list = defaultdict(list)
    for dir_path, dir_names, file_names in os.walk(path_base):
        dir_name = dir_path.split(os.sep)[-1]  # name of the current directory

        if dir_path.startswith(path_build) or dir_path in files_exclude:
            continue  # skip: folder for build output, and excluded folders

        if dir_name.startswith('.') or (os.sep + '.' in dir_path):
            continue  # hidden folders and sub-folders

        for file in file_names:
            (name, extension), path = os.path.splitext(file), os.path.join(dir_path, file)

            if path in files_exclude or extension in ('.pyc', 'pyx') or name.startswith('.'):
                continue  # skip: excluded files, pyc/pyx files, and hidden files

            path_full = os.path.abspath(path)

            action = 'copy'
            if extension in ('.py',):
                if not name.startswith('__') and path_full not in files_keep:
                    action = 'cythonize'
            elif not copy_others:
                continue  # if not copying other files, skip the file

            dst = path.replace(path_base, path_build)

            act_list[action].append((path, dst))

    for (src, dst) in act_list.get('copy', ()) + act_list.get('cythonize', ()):
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        shutil.copyfile(src, dst)

    target_cythonize = [dst for (src, dst) in act_list.get('cythonize', ())]

    n_parallel = os.cpu_count() or 8

    # python code -> c code
    cythonized = cythonize(target_cythonize, nthreads=n_parallel, language_level=3)

    # c code -> dynamic library file
    path_build_tmp = os.path.join(path_build, '.tmp')
    setup(ext_modules=cythonized, script_args=["build_ext", "-b", path_build, "-t", path_build_tmp, "-j", n_parallel])

    # clean up
    for c_module in cythonized:
        for c_file in c_module.sources:
            os.remove(c_file)
    for py_file in target_cythonize:
        os.remove(py_file)

    shutil.rmtree(path_build_tmp)

    print("\n\nSuccessfully finished building package to: ", path_build)


def main(*args, **kwargs):
    t = time.time()
    build(*args, **kwargs)
    t = time.time() - t
    print('Time consumed to build code: %.2f seconds.' % t)


if __name__ == '__main__':
    os.makedirs('build', exist_ok=True)
    shutil.rmtree('build')
    folder_name = os.getcwd().split(os.sep)[-1]
    folder_dist = os.path.join('/tmp/build/', folder_name)
    print('Building project to path:', folder_dist)
    shutil.rmtree(folder_dist, ignore_errors=True)

    main(
        base=None,  # use current directory by default
        dist=folder_dist,  # target directory for build files
        exclude=[__file__],  # exclude this file by default, this is a collection of files/folders to exclude
        keep=['./main.py'],  # source files keep as is and not converting to dynamic library
    )
    shutil.move(src=folder_dist, dst='./build')



__all__ = ('get_resource_dir', 'get_config_dir', 'get_current_module_dir', 'get_project_base_dir')

import os

_pjoin = os.path.join
_env = os.environ


def get_resource_dir():
    dir_cwd = os.getcwd()
    dir_resource = _env.get('DIR_RESOURCE', 'resource')
    return _pjoin(dir_cwd, dir_resource)


def get_config_dir():
    dir_config = _env.get('DIR_CONFIG')
    dir_resource = get_resource_dir()
    if dir_config is None or len(dir_config.strip()) == 0:
        dir_config = 'config'
    dir_config = _pjoin(dir_resource, dir_config)
    print(' ---> Using config dir:', dir_config)
    return dir_config


def get_current_module_dir(file_caller: str):
    dirs = file_caller.split(os.sep)
    return os.sep.join(dirs[:-1])


def get_project_base_dir(file_caller: str):
    dirs = file_caller.split(os.sep)

    dir_base = None
    for i in range(len(dirs)):
        dir_base = os.sep.join(dirs[:-1 - i])
        file_init = _pjoin(dir_base, '__init__.py')
        if not os.path.exists(file_init):
            break
    return dir_base



import logging
import os

from tornado import web, httpserver
from tornado.routing import HostMatches

from ..logger import LOG
from ..logger.logger import setup_logger
from ..settings import SETTINGS

setup_logger(
    logging.getLogger("tornado.access")
    , formatter_str='A> %(asctime)s> %(message)s'
    , module='access_%s' % (SETTINGS.config.get('APP_MODULE') or os.environ.get('APP_MODULE', 'default'))
)


def _load_handlers(name):
    """Load the (URL pattern, handler) tuples for each component."""
    mod = __import__(name, fromlist=['default_handlers'])
    handlers = []
    for url, handler in mod.default_handlers:

        if not url.startswith('/'):
            url = '/' + url
        handlers.append((url, handler))
    return handlers


class WebApplication(web.Application):
    def __init__(self, config: dict, *args, **kwargs):
        handlers = self.init_handlers(config)
        super().__init__(handlers=handlers, **config)
        self.http_server = httpserver.HTTPServer(self)

    @staticmethod
    def init_handlers(config: dict):
        settings = config.get('service', {})
        modules = settings.get('modules', [])
        handlers = []
        for m in modules:
            _handlers = _load_handlers(m)
            for h in _handlers:
                handlers.append(h)
                LOG.debug('Loaded module %s handler for API:\t%s' % h[::-1])

        return [
            (HostMatches('(.*)'), handlers)
        ]

    def start(self):
        service_settings = self.settings.get('service', {})

        port = service_settings.get('port', 80)
        # if overwrite port in param
        port = os.environ.get('port', port)

        num_process = int(service_settings.get('num_process', 1))
        LOG.info('Starting service with [%s] process at port [%s]...', num_process, port)
        self.http_server.bind(port)
        self.http_server.start(num_processes=num_process)
